import base
import os

from tap_tester import runner, connections

from base import FacebookBaseTest


class FacebookAttributionWindow(FacebookBaseTest):

    is_done = None

    @staticmethod
    def name():
        return "tap_tester_facebook_attribution_window"

    def streams_to_test(self):
        """ 'attribution window' is only supported for 'ads_insights' streams """

        # Fail the test when the JIRA card is done to allow stream to be re-added and tested
        if self.is_done is None:
            self.is_done = base.JIRA_CLIENT.get_status_category("TDL-24312") == 'done'
            self.assert_message = ("JIRA ticket has moved to done, re-add the "
                                   "ads_insights_hourly_advertiser stream to the test.")
        assert self.is_done != True, self.assert_message

        # return [stream for stream in self.expected_streams() if self.is_insight(stream)]
        return [stream for stream in self.expected_streams()
                if self.is_insight(stream)
                and stream != 'ads_insights_hourly_advertiser']

    def get_properties(self, original: bool = True):
        """Configuration properties required for the tap."""
        return_value = {
            'account_id': os.getenv('TAP_FACEBOOK_ACCOUNT_ID'),
            'start_date': self.start_date,
            'end_date': self.end_date,
            'insights_buffer_days': str(self.ATTRIBUTION_WINDOW)
        }
        if original:
            return return_value

        return_value["start_date"] = self.start_date
        return return_value

    def test_run(self):
        """
        For the test ad set up in facebook ads manager we see data
        on April 7th, start date is based on this data
        """
        # attrribution window = 7
        self.ATTRIBUTION_WINDOW = 7
        self.start_date = '2021-04-14T00:00:00Z'
        self.end_date = '2021-04-15T00:00:00Z'
        self.run_test(self.ATTRIBUTION_WINDOW, self.start_date, self.end_date)

        # attribution window = 28
        self.ATTRIBUTION_WINDOW = 28
        self.start_date = '2021-04-30T00:00:00Z'
        self.end_date = '2021-05-01T00:00:00Z'
        self.run_test(self.ATTRIBUTION_WINDOW, self.start_date, self.end_date)

        # attribution window = 1
        self.ATTRIBUTION_WINDOW = 1
        self.start_date = '2021-04-08T00:00:00Z'
        self.end_date = '2021-04-09T00:00:00Z'
        self.run_test(self.ATTRIBUTION_WINDOW, self.start_date, self.end_date)

    def run_test(self, attr_window, start_date, end_date):
        """
            Test to check the attribution window
        """

        expected_streams = self.streams_to_test()

        conn_id = connections.ensure_connection(self)

        # calculate start date with attribution window
        start_date_with_attribution_window = self.timedelta_formatted(
            start_date, days=-attr_window, date_format=self.START_DATE_FORMAT
        )

        # Run in check mode
        found_catalogs = self.run_and_verify_check_mode(conn_id)

        # Select only the expected streams tables
        catalog_entries = [ce for ce in found_catalogs if ce['tap_stream_id'] in expected_streams]
        self.perform_and_verify_table_and_field_selection(conn_id, catalog_entries, select_all_fields=True)

        # Run a sync job using orchestrator
        self.run_and_verify_sync(conn_id)
        sync_records = runner.get_records_from_target_output()

        expected_replication_keys = self.expected_replication_keys()

        for stream in expected_streams:
            with self.subTest(stream=stream):

                replication_key = next(iter(expected_replication_keys[stream]))

                # get records
                records = [record.get('data') for record in sync_records.get(stream).get('messages')]

                # check for the record is between attribution date and start date
                is_between = False

                for record in records:
                    replication_key_value = record.get(replication_key)

                    # Verify the sync records respect the attribution window
                    self.assertGreaterEqual(self.parse_date(replication_key_value), self.parse_date(start_date_with_attribution_window),
                                            msg="The record does not respect the attribution window.")

                    # verify if the record's bookmark value is between start date and attribution window
                    if self.parse_date(start_date_with_attribution_window) <= self.parse_date(replication_key_value) <= self.parse_date(start_date):
                        is_between = True

                    self.assertTrue(is_between)



"""
Test that with no fields selected for a stream all fields are still replicated
"""

from tap_tester.base_suite_tests.all_fields_test import AllFieldsTest
from base_new_frmwrk import FacebookBaseTest
import base


class FacebookAllFieldsTest(AllFieldsTest, FacebookBaseTest):
    """Test that with no fields selected for a stream all fields are still replicated"""

    is_done = None

    # https://jira.talendforge.org/browse/TDL-24424
    MISSING_FIELDS = {
        "ads_insights" : {
            'outbound_clicks',
            'conversions',
            'cost_per_action_type',
            'video_p100_watched_actions',
            'video_p75_watched_actions',
            'action_values',
            'video_30_sec_watched_actions',
            'canvas_avg_view_percent',
            'video_p50_watched_actions',
            'video_p25_watched_actions',
            'conversion_values',
            'unique_outbound_clicks',
            'canvas_avg_view_time',
            'cost_per_unique_action_type'
        },
        "campaigns" : {
            'adlabels'
        },
        "adsets" : {
            'adlabels',
            },
        "adcreative" : {
            'image_crops',
            'product_set_id',
            'url_tags',
            'applink_treatment',
            'object_id',
            'link_og_id',
            'template_url',
            'template_url_spec',
            'object_url',
            'link_url',
            'adlabels',
            'instagram_story_id'
        },
        "ads_insights_country": {
            'video_p75_watched_actions',
            'conversions',
            'conversion_values',
            'canvas_avg_view_percent',
            'action_values',
            'unique_outbound_clicks',
            'cost_per_unique_action_type',
            'outbound_clicks',
            'social_spend',
            'video_p50_watched_actions',
            'engagement_rate_ranking',
            'video_p25_watched_actions',
            'quality_ranking',
            'video_play_curve_actions',
            'video_30_sec_watched_actions',
            'canvas_avg_view_time',
            'cost_per_action_type',
            'video_p100_watched_actions',
            'conversion_rate_ranking'
        },
        "ads_insights_age_and_gender": {
            'video_p75_watched_actions',
            'conversions',
            'conversion_values',
            'canvas_avg_view_percent',
            'action_values',
            'unique_outbound_clicks',
            'cost_per_unique_action_type',
            'outbound_clicks',
            'social_spend',
            'video_p50_watched_actions',
            'engagement_rate_ranking',
            'video_p25_watched_actions',
            'quality_ranking',
            'video_play_curve_actions',
            'video_30_sec_watched_actions',
            'canvas_avg_view_time',
            'cost_per_action_type',
            'video_p100_watched_actions',
            'conversion_rate_ranking'
        },
        "ads_insights_dma": {
            'video_p75_watched_actions',
            'conversions',
            'cost_per_unique_click',
            'inline_link_click_ctr',
            'conversion_values',
            'canvas_avg_view_percent',
            'action_values',
            'unique_ctr',
            'unique_outbound_clicks',
            'unique_inline_link_clicks',
            'cost_per_unique_action_type',
            'outbound_clicks',
            'social_spend',
            'cost_per_unique_inline_link_click',
            'unique_link_clicks_ctr',
            'video_p50_watched_actions',
            'engagement_rate_ranking',
            'unique_inline_link_click_ctr',
            'video_p25_watched_actions',
            'quality_ranking',
            'cpp',
            'video_play_curve_actions',
            'canvas_avg_view_time',
            'video_30_sec_watched_actions',
            'cost_per_action_type',
            'video_p100_watched_actions',
            'conversion_rate_ranking'
        },
        "ads_insights_region": {
            'video_p75_watched_actions',
            'conversions',
            'conversion_values',
            'canvas_avg_view_percent',
            'action_values',
            'unique_outbound_clicks',
            'cost_per_unique_action_type',
            'outbound_clicks',
            'social_spend',
            'video_p50_watched_actions',
            'engagement_rate_ranking',
            'video_p25_watched_actions',
            'quality_ranking',
            'video_play_curve_actions',
            'video_30_sec_watched_actions',
            'canvas_avg_view_time',
            'cost_per_action_type',
            'video_p100_watched_actions',
            'conversion_rate_ranking'
        },
        "ads_insights_hourly_advertiser": {
            'conversions',
            'cost_per_estimated_ad_recallers',
            'cost_per_unique_click',
            'cost_per_unique_outbound_click',
            'frequency',
            'conversion_values',
            'canvas_avg_view_percent',
            'cost_per_conversion',
            'cost_per_thruplay',
            'action_values',
            'full_view_impressions',
            'place_page_name',
            'instant_experience_outbound_clicks',
            'cost_per_unique_action_type',
            'estimated_ad_recallers',
            'outbound_clicks',
            'social_spend',
            'cost_per_unique_inline_link_click',
            'instant_experience_clicks_to_start',
            'attribution_setting',
            'engagement_rate_ranking',
            'purchase_roas',
            'reach',
            'cost_per_outbound_click',
            'estimated_ad_recall_rate',
            'quality_ranking',
            'cpp',
            'catalog_segment_value',
            'canvas_avg_view_time',
            'cost_per_action_type',
            'outbound_clicks_ctr',
            'qualifying_question_qualify_answer_rate',
            'converted_product_quantity',
            'converted_product_value',
            'instant_experience_clicks_to_open',
            'conversion_rate_ranking'
        },
        "ads": {
            'recommendations',
            'adlabels'
        },
        "ads_insights_platform_and_device": {
            'video_p75_watched_actions',
            'conversions',
            'conversion_values',
            'canvas_avg_view_percent',
            'action_values',
            'unique_outbound_clicks',
            'cost_per_unique_action_type',
            'outbound_clicks',
            'social_spend',
            'video_p50_watched_actions',
            'engagement_rate_ranking',
            'video_p25_watched_actions',
            'quality_ranking',
            'video_play_curve_actions',
            'conversion_rate_ranking',
            'video_30_sec_watched_actions',
            'canvas_avg_view_time',
            'cost_per_action_type',
            'video_p100_watched_actions',
            'placement'
        }
    }


    @staticmethod
    def name():
        return "tt_facebook_all_fields_test"

    def streams_to_test(self):
        #return set(self.expected_metadata().keys())
        # Fail the test when the JIRA card is done to allow stream to be re-added and tested
        if self.is_done is None:
            self.is_done = base.JIRA_CLIENT.get_status_category("TDL-24312") == 'done'
            self.assert_message = ("JIRA ticket has moved to done, re-add the "
                                   "ads_insights_hourly_advertiser stream to the test.")
        assert self.is_done != True, self.assert_message

        return self.expected_metadata().keys() - {'ads_insights_hourly_advertiser'}



import os
import dateutil.parser
import datetime
from base_new_frmwrk import FacebookBaseTest
from tap_tester.base_suite_tests.table_reset_test import TableResetTest


class FacebookTableResetTest(TableResetTest, FacebookBaseTest):
    """tap-salesforce Table reset test implementation
    Currently tests only the stream with Incremental replication method"""

    @staticmethod
    def name():
        return "tt_facebook_table_reset"

    def streams_to_test(self):
        return self.expected_stream_names()

    @property
    def reset_stream(self):
        return ('ads_insights_dma')

    
    def calculated_states_by_stream(self, current_state):

        """        The following streams barely make the cut:

        campaigns "2021-02-09T18:17:30.000000Z"
                  "2021-02-09T16:24:58.000000Z"

        adsets    "2021-02-09T18:17:41.000000Z"
                  "2021-02-09T17:10:09.000000Z"

        leads     '2021-04-07T20:09:39+0000',
                  '2021-04-07T20:08:27+0000',
        """
        timedelta_by_stream = {stream: [0,0,0]  # {stream_name: [days, hours, minutes], ...}
                               for stream in self.expected_stream_names()}
        timedelta_by_stream['campaigns'] = [0, 1, 0]
        timedelta_by_stream['adsets'] = [0, 1, 0]
        timedelta_by_stream['leads'] = [0, 0 , 1]

        stream_to_calculated_state = {stream: "" for stream in current_state['bookmarks'].keys()}
        for stream, state in current_state['bookmarks'].items():
            state_key, state_value = next(iter(state.keys())), next(iter(state.values()))
            state_as_datetime = dateutil.parser.parse(state_value)
            days, hours, minutes = timedelta_by_stream[stream]
            calculated_state_as_datetime = state_as_datetime - datetime.timedelta(days=days, hours=hours, minutes=minutes)

            state_format = '%Y-%m-%dT00:00:00+00:00' if self.is_insight(stream) else '%Y-%m-%dT%H:%M:%S-00:00'
            calculated_state_formatted = datetime.datetime.strftime(calculated_state_as_datetime, state_format)

            stream_to_calculated_state[stream] = {state_key: calculated_state_formatted}

        return stream_to_calculated_state

    def manipulate_state(self,current_state):
        new_states = {'bookmarks': dict()}
        simulated_states = self.calculated_states_by_stream(current_state)
        for stream, new_state in simulated_states.items():
            new_states['bookmarks'][stream] = new_state
        return new_states



import test_client as tc
import time
import unittest

from datetime import datetime as dt

from tap_tester.base_suite_tests.pagination_test import PaginationTest
from tap_tester import connections, runner, menagerie, LOGGER

from base_new_frmwrk import FacebookBaseTest

fb_client = tc.TestClient()


class FacebookDiscoveryTest(PaginationTest, FacebookBaseTest):
    """Standard Pagination Test"""

    @staticmethod
    def name():
        return "tt_facebook_pagination"
    def streams_to_test(self):
        # TODO ads_insights empty for account, no post via API, spike on generating data
        return {'adcreative', 'ads', 'adsets', 'campaigns'}

    def setUp(self):  # pylint: disable=invalid-name
        """
        Setup for tests in this module.
        """
        if PaginationTest.synced_records and PaginationTest.record_count_by_stream:
            return

        # instantiate connection
        conn_id = connections.ensure_connection(self)

        # run check mode
        found_catalogs = self.run_and_verify_check_mode(conn_id)

        # table and field selection
        test_catalogs = [catalog for catalog in found_catalogs
                         if catalog.get('stream_name') in self.streams_to_test()]

        # non_selected_fields are none
        self.perform_and_verify_table_and_field_selection(conn_id, test_catalogs)

        # ensure there is enough data to paginate
        start_date_dt = self.parse_date(self.start_date)
        date_range = {'since': dt.strftime(start_date_dt, "%Y-%m-%d"),
                      'until': dt.strftime(dt.now(), "%Y-%m-%d")}

        for stream in self.streams_to_test():
            limit = self.expected_page_size(stream)
            response = fb_client.get_account_objects(stream, limit, date_range)

            number_of_records = len(response['data'])
            # TODO move "if" logic below to client method get_account_objects()
            if number_of_records >= limit and response.get('paging', {}).get('next'):
                continue  # stream is ready for test, no need for futher action

            LOGGER.info(f"Stream: {stream} - Record count is less than max page size: {limit}, "
                        "posting more records to setUp the PaginationTest")

            for i in range(limit - number_of_records + 1):
                post_response = fb_client.create_account_objects(stream)
                LOGGER.info(f"Posted {i + 1} new {stream}, new total: {number_of_records + i + 1}")

        # run initial sync
        PaginationTest.record_count_by_stream = self.run_and_verify_sync_mode(conn_id)
        PaginationTest.synced_records = runner.get_records_from_target_output()



import os

from tap_tester import connections, runner, menagerie

from base import FacebookBaseTest

class FacebookArchivedData(FacebookBaseTest):

    @staticmethod
    def name():
        return "tap_tester_facebook_archived_data"

    def streams_to_test(self):
        """include_deleted is supported for below streams only"""
        return ['ads', 'adsets', 'campaigns']

    def get_properties(self, original: bool = True):
        """Configuration properties required for the tap."""
        return_value = {
            'account_id': os.getenv('TAP_FACEBOOK_ACCOUNT_ID'),
            'start_date' : '2021-10-06T00:00:00Z',
            'end_date' : '2021-10-07T00:00:00Z',
            'insights_buffer_days': '1',
            'include_deleted': 'false'
        }
        if original:
            return return_value

        return_value["include_deleted"] = 'true'
        return return_value

    def test_run(self):
        '''
            Testing the archived data with 'include_deleted' parameter
        '''
        expected_streams = self.streams_to_test()

        ##########################################################################
        ### First Sync with include_deleted = false
        ##########################################################################

        # instantiate connection with the include_deleted = false
        conn_id_1 = connections.ensure_connection(self)

        # run check mode
        found_catalogs_1 = self.run_and_verify_check_mode(conn_id_1)

        # table and field selection
        test_catalogs_1_all_fields = [catalog for catalog in found_catalogs_1
                                      if catalog.get('tap_stream_id') in expected_streams]
        self.perform_and_verify_table_and_field_selection(conn_id_1, test_catalogs_1_all_fields, select_all_fields=True)

        # run initial sync
        record_count_by_stream_1 = self.run_and_verify_sync(conn_id_1)
        synced_records_1 = runner.get_records_from_target_output()

        ##########################################################################
        ### Second Sync with include_deleted = true
        ##########################################################################

        # create a new connection with the include_deleted = true
        conn_id_2 = connections.ensure_connection(self, original_properties=False)

        # run check mode
        found_catalogs_2 = self.run_and_verify_check_mode(conn_id_2)

        # table and field selection
        test_catalogs_2_all_fields = [catalog for catalog in found_catalogs_2
                                      if catalog.get('tap_stream_id') in expected_streams]
        self.perform_and_verify_table_and_field_selection(conn_id_2, test_catalogs_2_all_fields, select_all_fields=True)

        # run sync
        record_count_by_stream_2 = self.run_and_verify_sync(conn_id_2)
        synced_records_2 = runner.get_records_from_target_output()

        for stream in expected_streams:
            with self.subTest(stream=stream):

                # expected primary keys
                expected_primary_keys = self.expected_primary_keys()[stream]

                # collect information about count of records
                record_count_sync_1 = record_count_by_stream_1.get(stream, 0)
                record_count_sync_2 = record_count_by_stream_2.get(stream, 0)

                # collect list and set of primary keys for all the records
                primary_keys_list_1 = [tuple(message.get('data').get(expected_pk) for expected_pk in expected_primary_keys)
                                       for message in synced_records_1.get(stream).get('messages')
                                       if message.get('action') == 'upsert']
                primary_keys_list_2 = [tuple(message.get('data').get(expected_pk) for expected_pk in expected_primary_keys)
                                       for message in synced_records_2.get(stream).get('messages')
                                       if message.get('action') == 'upsert']
                primary_keys_sync_1 = set(primary_keys_list_1)
                primary_keys_sync_2 = set(primary_keys_list_2)

                # collect list of effective_status for all the records
                records_status_sync1 = [message.get('data').get('effective_status')
                                       for message in synced_records_1.get(stream).get('messages')
                                       if message.get('action') == 'upsert']
                records_status_sync2 = [message.get('data').get('effective_status')
                                       for message in synced_records_2.get(stream).get('messages')
                                       if message.get('action') == 'upsert']

                # Verifying that no ARCHIVED records are returned for sync 1
                self.assertNotIn('ARCHIVED', records_status_sync1)

                # Verifying that ARCHIVED records are returned for sync 2
                self.assertIn('ARCHIVED', records_status_sync2)

                # Verify the number of records replicated in sync 2 is greater than the number
                # of records replicated in sync 1
                self.assertGreater(record_count_sync_2, record_count_sync_1)

                # Verify the records replicated in sync 1 were also replicated in sync 2
                self.assertTrue(primary_keys_sync_1.issubset(primary_keys_sync_2))



import base
import os

from tap_tester import connections, runner, LOGGER

from base import FacebookBaseTest


class FacebookStartDateTest(FacebookBaseTest):

    is_done = None
    start_date_1 = ""
    start_date_2 = ""

    @staticmethod
    def name():
        return "tap_tester_facebook_start_date_test"

    def streams_to_test(self):
        # Fail the test when the JIRA card is done to allow stream to be re-added and tested
        if self.is_done is None:
            self.is_done = base.JIRA_CLIENT.get_status_category("TDL-24312") == 'done'
            self.assert_message = ("JIRA ticket has moved to done, re-add the "
                                   "ads_insights_hourly_advertiser stream to the test.")
        assert self.is_done != True, self.assert_message

        return self.expected_streams() - {'ads_insights_hourly_advertiser'}

    def test_run(self):
        """Instantiate start date according to the desired data set and run the test"""

        self.start_date_1 = '2021-04-07T00:00:00Z'
        self.start_date_2 = self.timedelta_formatted(
            self.start_date_1, days=2, date_format=self.START_DATE_FORMAT
        )

        self.start_date = self.start_date_1

        expected_streams = self.streams_to_test()

        ##########################################################################
        ### First Sync
        ##########################################################################

        # instantiate connection
        conn_id_1 = connections.ensure_connection(self)

        # run check mode
        found_catalogs_1 = self.run_and_verify_check_mode(conn_id_1)

        # table and field selection
        test_catalogs_1_all_fields = [catalog for catalog in found_catalogs_1
                                      if catalog.get('tap_stream_id') in expected_streams]
        self.perform_and_verify_table_and_field_selection(conn_id_1, test_catalogs_1_all_fields, select_all_fields=True)

        # run initial sync
        record_count_by_stream_1 = self.run_and_verify_sync(conn_id_1)
        synced_records_1 = runner.get_records_from_target_output()

        ##########################################################################
        ### Update START DATE Between Syncs
        ##########################################################################

        LOGGER.info("REPLICATION START DATE CHANGE: %s ===>>> %s ", self.start_date, self.start_date_2)
        self.start_date = self.start_date_2

        ##########################################################################
        ### Second Sync
        ##########################################################################

        # create a new connection with the new start_date
        conn_id_2 = connections.ensure_connection(self, original_properties=False)

        # run check mode
        found_catalogs_2 = self.run_and_verify_check_mode(conn_id_2)

        # table and field selection
        test_catalogs_2_all_fields = [catalog for catalog in found_catalogs_2
                                      if catalog.get('tap_stream_id') in expected_streams]
        self.perform_and_verify_table_and_field_selection(conn_id_2, test_catalogs_2_all_fields, select_all_fields=True)

        # run sync
        record_count_by_stream_2 = self.run_and_verify_sync(conn_id_2)
        synced_records_2 = runner.get_records_from_target_output()

        for stream in expected_streams:
            with self.subTest(stream=stream):

                # expected values
                expected_primary_keys = self.expected_primary_keys()[stream]
                expected_insights_buffer = -1 * int(self.get_properties()['insights_buffer_days'])
                expected_start_date_1 = self.timedelta_formatted(
                    self.start_date_1, days=expected_insights_buffer, date_format=self.START_DATE_FORMAT
                )
                expected_start_date_2 = self.timedelta_formatted(
                    self.start_date_2, days=expected_insights_buffer, date_format=self.START_DATE_FORMAT
                )

                # collect information for assertions from syncs 1 & 2 base on expected values
                record_count_sync_1 = record_count_by_stream_1.get(stream, 0)
                record_count_sync_2 = record_count_by_stream_2.get(stream, 0)
                primary_keys_list_1 = [tuple(message.get('data').get(expected_pk) for expected_pk in expected_primary_keys)
                                       for message in synced_records_1.get(stream).get('messages')
                                       if message.get('action') == 'upsert']
                primary_keys_list_2 = [tuple(message.get('data').get(expected_pk) for expected_pk in expected_primary_keys)
                                       for message in synced_records_2.get(stream).get('messages')
                                       if message.get('action') == 'upsert']
                primary_keys_sync_1 = set(primary_keys_list_1)
                primary_keys_sync_2 = set(primary_keys_list_2)

                if self.is_insight(stream):

                    # collect information specific to incremental streams from syncs 1 & 2
                    expected_replication_key = next(iter(self.expected_replication_keys().get(stream)))
                    replication_dates_1 =[row.get('data').get(expected_replication_key) for row in
                                          synced_records_1.get(stream, {'messages': []}).get('messages', [])
                                          if row.get('data')]
                    replication_dates_2 =[row.get('data').get(expected_replication_key) for row in
                                          synced_records_2.get(stream, {'messages': []}).get('messages', [])
                                          if row.get('data')]

                    # # Verify replication key is greater or equal to start_date for sync 1
                    for replication_date in replication_dates_1:
                        self.assertGreaterEqual(
                            self.parse_date(replication_date), self.parse_date(expected_start_date_1),
                                msg="Report pertains to a date prior to our start date.\n" +
                                "Sync start_date: {}\n".format(expected_start_date_1) +
                                "Record date: {} ".format(replication_date)
                        )

                    # Verify replication key is greater or equal to start_date for sync 2
                    for replication_date in replication_dates_2:
                        self.assertGreaterEqual(
                            self.parse_date(replication_date), self.parse_date(expected_start_date_2),
                                msg="Report pertains to a date prior to our start date.\n" +
                                "Sync start_date: {}\n".format(expected_start_date_2) +
                                "Record date: {} ".format(replication_date)
                        )

                    # Verify the number of records replicated in sync 1 is greater than the number
                    # of records replicated in sync 2
                    self.assertGreater(record_count_sync_1, record_count_sync_2)

                    # Verify the records replicated in sync 2 were also replicated in sync 1
                    self.assertTrue(primary_keys_sync_2.issubset(primary_keys_sync_1))

                else:

                    # Verify that the 2nd sync with a later start date replicates the same number of
                    # records as the 1st sync.
                    self.assertEqual(record_count_sync_2, record_count_sync_1)

                    # Verify by primary key the same records are replicated in the 1st and 2nd syncs
                    self.assertSetEqual(primary_keys_sync_1, primary_keys_sync_2)



"""
Test that with no fields selected for a stream automatic fields are still replicated
"""
import os

from tap_tester import runner, connections

from base import FacebookBaseTest


class FacebookAutomaticFields(FacebookBaseTest):
    """Test that with no fields selected for a stream automatic fields are still replicated"""

    @staticmethod
    def name():
        return "tap_tester_facebook_automatic_fields"

    def streams_to_test(self):
        return self.expected_streams()

    def get_properties(self, original: bool = True):
        """Configuration properties required for the tap."""
        return_value = {
            'account_id': os.getenv('TAP_FACEBOOK_ACCOUNT_ID'),
            'start_date' : '2021-04-08T00:00:00Z',
            'end_date' : '2021-04-08T00:00:00Z',
            'insights_buffer_days': '1'
        }
        if original:
            return return_value

        return_value["start_date"] = self.start_date
        return return_value


    def test_run(self):
        """
        Verify that for each stream you can get multiple pages of data
        when no fields are selected and only the automatic fields are replicated.

        PREREQUISITE
        For EACH stream add enough data that you surpass the limit of a single
        fetch of data.  For instance if you have a limit of 250 records ensure
        that 251 (or more) records have been posted for that stream.
        """

        expected_streams = self.streams_to_test()

        # instantiate connection
        conn_id = connections.ensure_connection(self)

        # run check mode
        found_catalogs = self.run_and_verify_check_mode(conn_id)

        # table and field selection
        test_catalogs_automatic_fields = [catalog for catalog in found_catalogs
                                          if catalog.get('stream_name') in expected_streams]

        self.perform_and_verify_table_and_field_selection(
            conn_id, test_catalogs_automatic_fields, select_all_fields=False,
        )

        # run initial sync
        record_count_by_stream = self.run_and_verify_sync(conn_id)
        synced_records = runner.get_records_from_target_output()

        for stream in expected_streams:
            with self.subTest(stream=stream):

                # expected values
                expected_keys = self.expected_automatic_fields().get(stream)

                # collect actual values
                data = synced_records.get(stream)
                record_messages_keys = [set(row['data'].keys()) for row in data['messages']]


                # Verify that you get some records for each stream
                self.assertGreater(
                    record_count_by_stream.get(stream, -1), 0,
                    msg="The number of records is not over the stream max limit")

                # Verify that only the automatic fields are sent to the target
                for actual_keys in record_messages_keys:
                    self.assertSetEqual(expected_keys, actual_keys)



import os
import random
import requests
import string

from tap_tester.logger import LOGGER


class TestClient():
    def __init__(self):
        self.base_url  = 'https://graph.facebook.com'
        self.api_version = 'v18.0'
        self.account_id = os.getenv('TAP_FACEBOOK_ACCOUNT_ID')
        self.access_token  = os.getenv('TAP_FACEBOOK_ACCESS_TOKEN')
        self.account_url = f"{self.base_url}/{self.api_version}/act_{self.account_id}"

        self.stream_endpoint_map = {'ads': '/ads',
                                    'adsets': '/adsets',
                                    'adcreative': '/adcreatives',
                                    'ads_insights': '/insights',  # GET only endpoint
                                    'campaigns': '/campaigns',
                                    'users': '/users',}

        self.campaign_special_ad_categories = ['NONE',
                                               'EMPLOYMENT',
                                               'HOUSING',
                                               'CREDIT',
                                               # 'ISSUES_ELECTIONS_POLITICS',  # acct unauthorized
                                               'ONLINE_GAMBLING_AND_GAMING']

        self.adset_billing_events = ['APP_INSTALLS',
                                     'CLICKS',
                                     'IMPRESSIONS',
                                     'LINK_CLICKS',
                                     'NONE',
                                     'OFFER_CLAIMS',
                                     'PAGE_LIKES',
                                     'POST_ENGAGEMENT',
                                     'THRUPLAY',
                                     'PURCHASE',
                                     'LISTING_INTERACTION']

        self.adset_optimization_goals = ['NONE',
                                         'APP_INSTALLS',
                                         'AD_RECALL_LIFT',
                                         'ENGAGED_USERS',
                                         'EVENT_RESPONSES',
                                         'IMPRESSIONS',
                                         'LEAD_GENERATION',
                                         'QUALITY_LEAD',
                                         'LINK_CLICKS',
                                         'OFFSITE_CONVERSIONS',
                                         'PAGE_LIKES',
                                         'POST_ENGAGEMENT',
                                         'QUALITY_CALL',
                                         'REACH',
                                         'LANDING_PAGE_VIEWS',
                                         'VISIT_INSTAGRAM_PROFILE',
                                         'VALUE',
                                         'THRUPLAY',
                                         'DERIVED_EVENTS',
                                         'APP_INSTALLS_AND_OFFSITE_CONVERSIONS',
                                         'CONVERSATIONS',
                                         'IN_APP_VALUE',
                                         'MESSAGING_PURCHASE_CONVERSION',
                                         'MESSAGING_APPOINTMENT_CONVERSION',
                                         'SUBSCRIBERS',
                                         'REMINDERS_SET']

        # valid and verified objectives listed below, other objectives should be re-mapped to these
        self.campaign_objectives = ['OUTCOME_APP_PROMOTION',
                                    'OUTCOME_AWARENESS',
                                    'OUTCOME_ENGAGEMENT',
                                    'OUTCOME_LEADS',
                                    'OUTCOME_SALES',
                                    'OUTCOME_TRAFFIC']

    def get_account_objects(self, stream, limit, time_range):
        # time_range defines query start and end dates and should match tap config
        assert stream in  self.stream_endpoint_map.keys(), \
            f'Endpoint undefined for specified stream: {stream}'
        endpoint = self.stream_endpoint_map[stream]
        url = self.account_url + endpoint
        params = {'access_token': self.access_token,
                  'limit': limit,
                  'time_range': str({'since': time_range['since'], 'until': time_range['until']})}
        LOGGER.info(f"Getting url: {url}")
        response = requests.get(url, params)
        response.raise_for_status()
        LOGGER.info(f"Returning get response: {response}")
        return response.json()

    def create_account_objects(self, stream):
        assert stream in  self.stream_endpoint_map.keys(), \
            f'Endpoint undefined for specified stream: {stream}'
        endpoint = self.stream_endpoint_map[stream]
        url = self.account_url + endpoint
        LOGGER.info(f"Posting to url: {url}")
        params = self.generate_post_params(stream)
        response = requests.post(url, params)
        response.raise_for_status()
        LOGGER.info(f"Returning post response: {response}")
        return response

    def generate_post_params(self, stream):
        if stream == 'adcreative':
            params = {
                'access_token': self.access_token,
                'name': ''.join(random.choices(string.ascii_letters + string.digits, k=18)),
                'object_story_spec': str({'page_id': '453760455405317',
                                          'link_data': {'link': 'http://fb.me'}})}
            return params

        elif stream == 'ads':
            params = {
                'access_token': self.access_token,
                'name': ''.join(random.choices(string.ascii_letters + string.digits, k=17)),
                # adset is bound to parent campaign_objective, can cause errors posting new ads
                #   as certain objectives have different requirements. 50 ads per adset max
                #   adset below can be found under campaign: 120203395323750059
                'adset_id': 120203403135680059,
                'creative': str({'creative_id': 23843561378450058}),
                'status': "PAUSED"}
            return params

        elif stream == 'adsets':
            # TODO In order to randomize optimization_goal and billing_event the campaign_id
            #   would need to be examined to determine which goals were supported.  Then an option
            #   could be selected from the available billing events supported by that goal.
            params = {
                'access_token': self.access_token,
                'name': ''.join(random.choices(string.ascii_letters + string.digits, k=16)),
                'optimization_goal': 'REACH',
                'billing_event': 'IMPRESSIONS',
                'bid_amount': 2,  # TODO random?
                'daily_budget': 1000, # TODO random? tie to parent campaign?
                'campaign_id': 120203241386960059,
                'targeting': str({'geo_locations': {'countries': ["US"]},
                                  'facebook_positions': ["feed"]}),
                'status': "PAUSED",
                'promoted_object': str({'page_id': '453760455405317'})}
            return params

        elif stream == 'campaigns':
            params = {  # generate a campaign with random name, ojbective, and ad category
                'access_token': self.access_token,
                'name': ''.join(random.choices(string.ascii_letters + string.digits, k=15)),
                'objective': random.choice(self.campaign_objectives),
                'special_ad_categories': random.choice(self.campaign_special_ad_categories)}
            return params

        else:
            assert False, f"Post params for stream {stream} not implemented / supported"


    # Ad Insights TODO
    # endpoint is "GET" only.  We cannot post fake insights data for test. As of Oct 27, 2023
    # data lists for original 3 AdSet Ids, and Ads account as a whole are empty.
    #   1 - Can we run enough ads to get enough data to paginate?
    #   2 - Can we interact with our own ads?
    # if 1 or 2 == True then use setUp to conditionally test ads_insights if there is enough data



import os
import unittest

from tap_tester import runner, connections, menagerie

from base import FacebookBaseTest

class FacebookInvalidAttributionWindowInt(FacebookBaseTest):

    @staticmethod
    def name():
        return "tt_facebook_invalid_window_int"

    @staticmethod
    def streams_to_test():
        return []

    def get_properties(self, original: bool = True):
        """Configuration properties required for the tap."""
        return_value = {
            'account_id': os.getenv('TAP_FACEBOOK_ACCOUNT_ID'),
            'start_date' : '2019-07-24T00:00:00Z',
            'end_date' : '2019-07-26T00:00:00Z',
            'insights_buffer_days': self.ATTRIBUTION_WINDOW,
        }
        if original:
            return return_value

        return_value["start_date"] = self.start_date
        return return_value

    def test_run(self):
        self.ATTRIBUTION_WINDOW = '10' # set attribution window other than 1, 7 or 28
        self.run_test()

    def run_test(self):
        """
        Test to verify that the error is raise when passing attribution window other than 1, 7 or 28
        """
        # create connection
        conn_id = connections.ensure_connection(self)
        # run check mode
        check_job_name = runner.run_check_mode(self, conn_id)
        # get exit status
        exit_status = menagerie.get_exit_status(conn_id, check_job_name)
        # get discovery error message
        discovery_error_message = exit_status.get('discovery_error_message')
        # validate the error message
        self.assertEquals(discovery_error_message, "The attribution window must be 1, 7 or 28.")
        self.assertIsNone(exit_status.get('target_exit_status'))
        self.assertIsNone(exit_status.get('tap_exit_status'))


class FacebookInvalidAttributionWindowStr(FacebookInvalidAttributionWindowInt):

    @staticmethod
    def name():
        return "tt_facebook_invalid_window_str"
    
    @unittest.skip("BUG: TDL-18569")
    def test_run(self):
        self.ATTRIBUTION_WINDOW = 'something'
        self.run_test()

    # BUG : TDL-18569 created to standarize the error message for sting values for attribution window



import base
import datetime
import dateutil.parser
import os
import pytz

from tap_tester import runner, menagerie, connections

from base import FacebookBaseTest


class FacebookBookmarks(FacebookBaseTest):

    is_done = None

    @staticmethod
    def name():
        return "tap_tester_facebook_bookmarks"

    def streams_to_test(self):
        # Fail the test when the JIRA card is done to allow stream to be re-added and tested
        if self.is_done is None:
            self.is_done = base.JIRA_CLIENT.get_status_category("TDL-24312") == 'done'
            self.assert_message = ("JIRA ticket has moved to done, re-add the "
                                   "ads_insights_hourly_advertiser stream to the test.")
        assert self.is_done != True, self.assert_message

        return self.expected_streams() - {'ads_insights_hourly_advertiser'}

    @staticmethod
    def convert_state_to_utc(date_str):
        """
        Convert a saved bookmark value of the form '2020-08-25T13:17:36-07:00' to
        a string formatted utc datetime,
        in order to compare aginast json formatted datetime values
        """
        date_object = dateutil.parser.parse(date_str)
        date_object_utc = date_object.astimezone(tz=pytz.UTC)
        return datetime.datetime.strftime(date_object_utc, "%Y-%m-%dT%H:%M:%SZ")

    def calculated_states_by_stream(self, current_state):
        """
        Look at the bookmarks from a previous sync and set a new bookmark
        value based off timedelta expectations. This ensures the subsequent sync will replicate
        at least 1 record but, fewer records than the previous sync.

        Sufficient test data is required for this test to cover a given stream.
        An incrmeental replication stream must have at least two records with
        replication keys that differ by more than the lookback window.

        If the test data is changed in the future this will break expectations for this test.

        The following streams barely make the cut:

        campaigns "2021-02-09T18:17:30.000000Z"
                  "2021-02-09T16:24:58.000000Z"

        adsets    "2021-02-09T18:17:41.000000Z"
                  "2021-02-09T17:10:09.000000Z"

        leads     '2021-04-07T20:09:39+0000',
                  '2021-04-07T20:08:27+0000',
        """
        # TODO We want to move this bookmark back by some amount for insgihts streams but
        # cannot do that unless we have at least 3 days of data. Currently we have 2.
        timedelta_by_stream = {stream: [0,0,0]  # {stream_name: [days, hours, minutes], ...}
                               for stream in self.expected_streams()}
        timedelta_by_stream['campaigns'] = [0, 1, 0]
        timedelta_by_stream['adsets'] = [0, 1, 0]
        timedelta_by_stream['leads'] = [0, 0 , 1]

        stream_to_calculated_state = {stream: "" for stream in current_state['bookmarks'].keys()}
        for stream, state in current_state['bookmarks'].items():
            state_key, state_value = next(iter(state.keys())), next(iter(state.values()))
            state_as_datetime = dateutil.parser.parse(state_value)

            days, hours, minutes = timedelta_by_stream[stream]
            calculated_state_as_datetime = state_as_datetime - datetime.timedelta(days=days, hours=hours, minutes=minutes)

            state_format = '%Y-%m-%dT00:00:00+00:00' if self.is_insight(stream) else '%Y-%m-%dT%H:%M:%S-00:00'
            calculated_state_formatted = datetime.datetime.strftime(calculated_state_as_datetime, state_format)

            stream_to_calculated_state[stream] = {state_key: calculated_state_formatted}

        return stream_to_calculated_state

    # function for verifying the date format
    def is_expected_date_format(self, date):
        try:
            # parse date
            datetime.datetime.strptime(date, "%Y-%m-%dT%H:%M:%S.%fZ")
        except ValueError:
            # return False if date is in not expected format
            return False
        # return True in case of no error
        return True

    def test_run(self):
        expected_streams = self.streams_to_test()
        non_insight_streams = {stream for stream in expected_streams if not self.is_insight(stream)}
        insight_streams = {stream for stream in expected_streams if self.is_insight(stream)}

        # Testing against ads insights objects
        self.start_date = self.get_properties()['start_date']
        self.end_date = self.get_properties()['end_date']
        self.bookmarks_test(insight_streams)

        # Testing against core objects
        self.end_date = '2021-02-09T00:00:00Z'
        self.bookmarks_test(non_insight_streams)


    def bookmarks_test(self, expected_streams):
        """A Parametrized Bookmarks Test"""
        expected_replication_keys = self.expected_replication_keys()
        expected_replication_methods = self.expected_replication_method()
        expected_insights_buffer = -1 * int(self.get_properties()['insights_buffer_days'])  # lookback window

        ##########################################################################
        ### First Sync
        ##########################################################################

        conn_id = connections.ensure_connection(self, original_properties=False)

        # Run in check mode
        found_catalogs = self.run_and_verify_check_mode(conn_id)

        # Select only the expected streams tables
        catalog_entries = [ce for ce in found_catalogs if ce['tap_stream_id'] in expected_streams]
        self.perform_and_verify_table_and_field_selection(conn_id, catalog_entries, select_all_fields=True)

        # Run a sync job using orchestrator
        first_sync_record_count = self.run_and_verify_sync(conn_id)
        first_sync_records = runner.get_records_from_target_output()
        first_sync_bookmarks = menagerie.get_state(conn_id)

        ##########################################################################
        ### Update State Between Syncs
        ##########################################################################

        new_states = {'bookmarks': dict()}
        simulated_states = self.calculated_states_by_stream(first_sync_bookmarks)
        for stream, new_state in simulated_states.items():
            new_states['bookmarks'][stream] = new_state
        menagerie.set_state(conn_id, new_states)

        ##########################################################################
        ### Second Sync
        ##########################################################################

        second_sync_record_count = self.run_and_verify_sync(conn_id)
        second_sync_records = runner.get_records_from_target_output()
        second_sync_bookmarks = menagerie.get_state(conn_id)

        ##########################################################################
        ### Test By Stream
        ##########################################################################

        for stream in expected_streams:
            with self.subTest(stream=stream):

                # expected values
                expected_replication_method = expected_replication_methods[stream]

                # collect information for assertions from syncs 1 & 2 base on expected values
                first_sync_count = first_sync_record_count.get(stream, 0)
                second_sync_count = second_sync_record_count.get(stream, 0)
                first_sync_messages = [record.get('data') for record in
                                       first_sync_records.get(stream).get('messages')
                                       if record.get('action') == 'upsert']
                second_sync_messages = [record.get('data') for record in
                                        second_sync_records.get(stream).get('messages')
                                        if record.get('action') == 'upsert']
                first_bookmark_key_value = first_sync_bookmarks.get('bookmarks', {stream: None}).get(stream)
                second_bookmark_key_value = second_sync_bookmarks.get('bookmarks', {stream: None}).get(stream)


                if expected_replication_method == self.INCREMENTAL:


                    # collect information specific to incremental streams from syncs 1 & 2
                    replication_key = next(iter(expected_replication_keys[stream]))
                    first_bookmark_value = first_bookmark_key_value.get(replication_key)
                    second_bookmark_value = second_bookmark_key_value.get(replication_key)
                    first_bookmark_value_utc = self.convert_state_to_utc(first_bookmark_value)
                    second_bookmark_value_utc = self.convert_state_to_utc(second_bookmark_value)
                    simulated_bookmark_value = new_states['bookmarks'][stream][replication_key]
                    simulated_bookmark_minus_lookback = self.timedelta_formatted(
                        simulated_bookmark_value, days=expected_insights_buffer,
                        date_format=self.BOOKMARK_COMPARISON_FORMAT
                    ) if self.is_insight(stream) else simulated_bookmark_value


                    # Verify the first sync sets a bookmark of the expected form
                    self.assertIsNotNone(first_bookmark_key_value)
                    self.assertIsNotNone(first_bookmark_key_value.get(replication_key))

                    # Verify the second sync sets a bookmark of the expected form
                    self.assertIsNotNone(second_bookmark_key_value)
                    self.assertIsNotNone(second_bookmark_key_value.get(replication_key))

                    # Verify the second sync bookmark is Equal to the first sync bookmark
                    self.assertEqual(second_bookmark_value, first_bookmark_value) # assumes no changes to data during test


                    for record in second_sync_messages:
                        # for "ads_insights_age_and_gender" and "ads_insights_hourly_advertiser"
                        # verify that the "date_start" and "date_stop" is in expected format
                        if stream in ["ads_insights_age_and_gender", "ads_insights_hourly_advertiser"]:
                            date_start = record.get("date_start")
                            self.assertTrue(self.is_expected_date_format(date_start))
                            date_stop = record.get("date_stop")
                            self.assertTrue(self.is_expected_date_format(date_stop))

                        # Verify the second sync records respect the previous (simulated) bookmark value
                        replication_key_value = record.get(replication_key)
                        self.assertGreaterEqual(replication_key_value, simulated_bookmark_minus_lookback,
                                                msg="Second sync records do not repect the previous bookmark.")

                        # Verify the second sync bookmark value is the max replication key value for a given stream
                        self.assertLessEqual(
                            replication_key_value, second_bookmark_value_utc,
                            msg="Second sync bookmark was set incorrectly, a record with a greater replication-key value was synced."
                        )

                    for record in first_sync_messages:
                        # for "ads_insights_age_and_gender" and "ads_insights_hourly_advertiser"
                        # verify that the "date_start" and "date_stop" is in expected format
                        if stream in ["ads_insights_age_and_gender", "ads_insights_hourly_advertiser"]:
                            date_start = record.get("date_start")
                            self.assertTrue(self.is_expected_date_format(date_start))
                            date_stop = record.get("date_stop")
                            self.assertTrue(self.is_expected_date_format(date_stop))

                        # Verify the first sync bookmark value is the max replication key value for a given stream
                        replication_key_value = record.get(replication_key)
                        self.assertLessEqual(
                            replication_key_value, first_bookmark_value_utc,
                            msg="First sync bookmark was set incorrectly, a record with a greater replication-key value was synced."
                        )


                    # Verify the number of records in the 2nd sync is less then the first
                    self.assertLess(second_sync_count, first_sync_count)


                elif expected_replication_method == self.FULL_TABLE:


                    # Verify the syncs do not set a bookmark for full table streams
                    self.assertIsNone(first_bookmark_key_value)
                    self.assertIsNone(second_bookmark_key_value)

                    # Verify the number of records in the second sync is the same as the first
                    self.assertEqual(second_sync_count, first_sync_count)


                else:


                    raise NotImplementedError(
                        "INVALID EXPECTATIONS\t\tSTREAM: {} REPLICATION_METHOD: {}".format(stream, expected_replication_method)
                    )


                # Verify at least 1 record was replicated in the second sync
                self.assertGreater(second_sync_count, 0, msg="We are not fully testing bookmarking for {}".format(stream))



import unittest
from tap_tester.base_suite_tests.discovery_test import DiscoveryTest

from base_new_frmwrk import FacebookBaseTest


class FacebookDiscoveryTest(DiscoveryTest, FacebookBaseTest):
    """Standard Discovery Test"""

    @staticmethod
    def name():
        return "tt_facebook_discovery"
    def streams_to_test(self):
        return self.expected_stream_names()



import os
from functools import reduce

from tap_tester import connections, menagerie, runner, LOGGER

from base import FacebookBaseTest


class FacebookFieldSelection(FacebookBaseTest):  # TODO use base.py, determine if test is needed

    @staticmethod
    def name():
        return "tap_tester_facebook_field_selection"

    @staticmethod
    def streams_to_test():
        return {
            'ads',
            'adcreative',
            'adsets',
            'campaigns',
            'ads_insights',
            'ads_insights_age_and_gender',
            'ads_insights_country',
            'ads_insights_platform_and_device',
            'ads_insights_region',
            'ads_insights_dma',
            "ads_insights_hourly_advertiser",
            #'leads',
        }

    @staticmethod
    def expected_pks():
        return {
            "ads" :                             {"id", "updated_time"},
            "adcreative" :                      {"id"},
            "adsets" :                          {"id", "updated_time"},
            "campaigns" :                       {"id"},
            "ads_insights" :                    {"campaign_id", "adset_id", "ad_id", "date_start"},
            "ads_insights_age_and_gender" :     {"campaign_id", "adset_id", "ad_id", "date_start", "age", "gender"},
            "ads_insights_country" :            {"campaign_id", "adset_id", "ad_id", "date_start", "country"},
            "ads_insights_platform_and_device": {"campaign_id", "adset_id", "ad_id", "date_start", "publisher_platform", "platform_position", "impression_device"},
            "ads_insights_region" :             {"campaign_id", "adset_id", "ad_id", "date_start"},
            "ads_insights_dma" :                {"campaign_id", "adset_id", "ad_id", "date_start"},
            "ads_insights_hourly_advertiser":   {"campaign_id", "adset_id", "ad_id", "date_start", "hourly_stats_aggregated_by_advertiser_time_zone"},
            #"leads" :                           {"id"},
        }

    def expected_automatic_fields(self):
        return_value = self.expected_pks()
        return_value["campaigns"].add("updated_time")
        return return_value

    def get_properties(self):  # pylint: disable=arguments-differ
        return {'start_date' : '2015-03-15T00:00:00Z',
                'account_id': os.getenv('TAP_FACEBOOK_ACCOUNT_ID'),
                'end_date': '2015-03-16T00:00:00+00:00',
                'insights_buffer_days': '1'
        }

    def test_run(self):

        expected_streams = self.streams_to_test()

        conn_id = connections.ensure_connection(self)

        # run in check mode
        check_job_name = runner.run_check_mode(self, conn_id)

        # verify check exit codes
        exit_status = menagerie.get_exit_status(conn_id, check_job_name)
        menagerie.verify_check_exit_status(self, exit_status, check_job_name)

        found_catalogs = menagerie.get_catalogs(conn_id)
        self.assertGreater(len(found_catalogs), 0, msg="unable to locate schemas for connection {}".format(conn_id))

        found_catalog_names = set(map(lambda c: c['tap_stream_id'], found_catalogs))

        diff = expected_streams.symmetric_difference( found_catalog_names )
        self.assertEqual(len(diff), 0, msg="discovered schemas do not match: {}".format(diff))
        LOGGER.info("discovered schemas are kosher")

        all_excluded_fields = {}
        # select all catalogs
        for c in found_catalogs:
            if c['stream_name'] == 'ads':
                continue

            discovered_schema = menagerie.get_annotated_schema(conn_id, c['stream_id'])['annotated-schema']
            all_excluded_fields[c['stream_name']] = list(set(discovered_schema.keys()) - self.expected_automatic_fields().get(c['stream_name'], set()))[:5]
            connections.select_catalog_and_fields_via_metadata(
                conn_id,
                c,
                discovered_schema,
                non_selected_fields=all_excluded_fields[c['stream_name']])

        # clear state
        menagerie.set_state(conn_id, {})

        sync_job_name = runner.run_sync_mode(self, conn_id)

        # verify tap and target exit codes
        exit_status = menagerie.get_exit_status(conn_id, sync_job_name)
        menagerie.verify_sync_exit_status(self, exit_status, sync_job_name)

        # This should be validating the the PKs are written in each record
        record_count_by_stream = runner.examine_target_output_file(self, conn_id, expected_streams, self.expected_pks())
        replicated_row_count =  reduce(lambda accum,c : accum + c, record_count_by_stream.values())
        self.assertGreater(replicated_row_count, 0, msg="failed to replicate any data: {}".format(record_count_by_stream))
        print("total replicated row count: {}".format(replicated_row_count))

        synced_records = runner.get_records_from_target_output()
        self.assertTrue('ads' not in synced_records.keys())
        for stream_name, data in synced_records.items():
            record_messages = [set(row['data'].keys()) for row in data['messages']]
            for record_keys in record_messages:
                # The intersection should be empty
                self.assertFalse(record_keys.intersection(all_excluded_fields[stream_name]))
